/*
 * Liquid-filled tubes, October 13, 2007
 *
 * Each cylinder-drawing command accepts the centers of the bottom and
 * top faces, the tube's radius, and the depth of liquid, and assumes
 * the camera looks at the "top" end. There are few sanity checks
 * (fluid depth positive, less than tube height), and each command
 * sets the fill color to Neutral().
 *
 * //// User commands defined in this file ////
 * // cylinders
 * void  tube_vert(P ctr_bot, P ctr_top, double rad, double depth);
 * void tube_horiz(P ctr_bot, P ctr_top, double rad, double depth);
 *
 * void tube_cone(P apex, P ctr, double rad, double depth);
 * void tube_sphere(P ctr, double rad, double depth);
 */

#include "epix.h"

namespace ePiX {

  // construct a frame to calculate "horizon" of a cylinder from vpt
  frame cyl_coords(const P& axis1, const P& axis2, const P& vpt)
    {
      // sky() is perp to axis, points to vpt, eye() = axis
      return frame(E_1, vpt - axis1, axis2 - axis1);
    }

  // draw a cylinder as a pair of fillable paths
  void fill_v_cyl(const P& ctr_bot, const P& ctr_top, double rad)
    {
      frame basis(cyl_coords(ctr_bot, ctr_top, cam().viewpt()));

      // perp from viewpt to cylinder axis
      P to_axis((cam().viewpt() - ctr_bot) %= basis.eye());
      double dist(norm(to_axis));

      // get sin/cos of angle from sky() to horizon
      const double x(rad/dist);
      const double y(sqrt((dist-rad)*(dist+rad))/dist);

      const double th(Asin(x));
      const double qtr(0.25*full_turn());

      // now we're ready to draw...

      // visible arc on bottom
      path lateral(ctr_bot, rad*basis.sky(), rad*basis.sea(), -qtr+th, qtr-th);

      // up one side
      lateral += path(ctr_bot + rad*(x*basis.sky() + y*basis.sea()),
		      ctr_top + rad*(x*basis.sky() + y*basis.sea()));

      // front curve of top surface
      lateral += path(ctr_top, rad*basis.sky(), rad*basis.sea(),
		      qtr-th, -qtr+th);

      // down the other side
      lateral += path(ctr_top + rad*(x*basis.sky() - y*basis.sea()),
		      ctr_bot + rad*(x*basis.sky() - y*basis.sea()));

      lateral.close().fill().draw();

      // re-draw top surface
      circle(ctr_top, rad, ctr_top - ctr_bot);
    }

  //// User-level command ////
  void tube_vert(const P& ctr_bot, const P& ctr_top, double rad, double depth)
    {
      P axis(ctr_top - ctr_bot);
      double len(norm(axis));

      if(depth < 0)
	depth = 0;

      else if (len < depth)
	depth = len;

      axis *= 1.0/len; // normalize

      // draw fluid, using current fill color
      if (0 < depth)
	fill_v_cyl(ctr_bot, ctr_bot + depth*axis, rad);

      // draw tube
      fill(Neutral()); // N.B. resets fill color
      fill_v_cyl(ctr_bot + depth*axis, ctr_top, rad);
    }


  // internal command; may assume 0 <= depth <= 2*rad
  void fill_h_cyl(const P& ctr_bot, const P& ctr_top, double rad, double depth)
    {
      // check for horizontality -- magic number 0.001
      if (0.001 < fabs(ctr_bot.x3() - ctr_top.x3()))
	return;

      // angle from vert down to fluid surface
      double th0(Acos(1 - depth/rad));

      // "horizon" coordinates (sky() on edge of cylinder)
      frame basis(cyl_coords(ctr_bot, ctr_top, cam().viewpt()));


      // Cartesian/vertical coordinates (perp parallel to liquid surface)
      P axis(ctr_top - ctr_bot);
      axis *= 1.0/norm(axis);
      P perp(axis*E_3);

      // calculate angle to horizon
      P to_axis((cam().viewpt() - ctr_bot) %= axis);
      double dist(norm(to_axis));

      const double x(rad/dist);
      const double y(sqrt((dist-rad)*(dist+rad))/dist);

      // express th in Cartesian/vertical coords
      const double th(Asin(x) + Asin(basis.sky()|E_3));

      // and draw...

      // visible arc on bottom
      path lateral(ctr_bot, -rad*E_3, rad*perp, th, th0);

      // front edge of surface
      lateral += path(ctr_bot + rad*(-Cos(th0)*E_3 + Sin(th0)*perp),
		      ctr_top + rad*(-Cos(th0)*E_3 + Sin(th0)*perp));

      // front arc on top face
      lateral += path(ctr_top, -rad*E_3, rad*perp, th0, th);

      // bottom horizon of tube
      lateral+=path(ctr_top + rad*(-Cos(th)*E_3 + Sin(th)*perp),
		    ctr_bot + rad*(-Cos(th)*E_3 + Sin(th)*perp));

      lateral.close().fill().draw();

      // "top end" of liquid
      path top_cap(ctr_top, -rad*E_3, rad*perp, -th0, th0);
      top_cap.close().fill().draw();

      // surface of liquid
      quad(ctr_bot + rad*(-Cos(th0)*E_3 + Sin(th0)*perp),
	   ctr_top + rad*(-Cos(th0)*E_3 + Sin(th0)*perp),
	   ctr_top + rad*(-Cos(th0)*E_3 - Sin(th0)*perp),
	   ctr_bot + rad*(-Cos(th0)*E_3 - Sin(th0)*perp));
    }

  //// User command ////
  void tube_horiz(const P& ctr_bot, const P& ctr_top, double rad, double depth)
    {
      P axis(ctr_top - ctr_bot);
      axis *= 1.0/norm(axis);

      // check for sane depth
      if (depth <= 0)
	depth = 0;

      else if (2*rad < depth)
	depth = 2*rad;

      if (depth < 2*rad)
	{
	  // draw far ("bottom") end of tube
	  nofill();
	  circle(ctr_bot, rad, ctr_top - ctr_bot);
	}

      if (0 < depth && depth < 2*rad)
	{
	  // fluid
	  fill();
	  fill_h_cyl(ctr_bot, ctr_top, rad, depth);
	}

      // and the tube
      if (depth < 2*rad)
	fill(Neutral()); // N.B. resets fill color

      fill_v_cyl(ctr_bot, ctr_top, rad); // N.B. fill_v_cyl
    }


  // draw a right circular cone as a pair of fillable paths
  void fill_cone(const P& apex, const P& ctr, double rad)
    {
      frame basis(cyl_coords(apex, ctr, cam().viewpt()));

      // perp from viewpt to cylinder axis
      P to_axis((cam().viewpt() - apex) %= basis.eye());
      double dist(norm(to_axis));

      // get sin/cos of angle from sky() to horizon
      const double x(rad/dist);
      const double y(sqrt((dist-rad)*(dist+rad))/dist);

      const double th(Asin(x));
      const double qtr(0.25*full_turn());

      // draw...
      path lateral(ctr, rad*basis.sky(), rad*basis.sea(), -qtr+th, qtr-th);
      lateral.pt(apex).close().fill().draw();

      /*
      triangle(apex,
	       ctr + rad*(x*basis.sky() + y*basis.sea()),
	       ctr + rad*(x*basis.sky() - y*basis.sea()));
      */

      // top surface
      circle(ctr, rad, ctr - apex);
    }

  //// User-level command ////
  void tube_cone(const P& apex, const P& ctr, double rad, double depth)
    {
      P axis(ctr - apex);
      double len(norm(axis));

      if(depth < 0)
	depth = 0;

      else if (len < depth)
	depth = len;

      axis *= 1.0/len; // normalize

      // draw fluid, using current fill color
      if (0 < depth)
	fill_cone(apex, apex + depth*axis, rad*depth/len);

      // draw tube
      fill(Neutral()); // N.B. resets fill color
      fill_cone(apex, ctr, rad);
    }

  void tube_sphere(const P& ctr, double rad, double depth)
    {
      Sphere S(ctr, rad);
      if (depth < 0)
	depth = 0;

      else if (2*rad < depth)
	depth = 2*rad;

      // will draw fluid by clipping sphere horizon
      clip_face(ctr+(depth-rad)*E_3, -E_3);
      S.draw();
      clip_restore();

      // fluid surface
      if (0 < depth && depth < 2*rad)
	circle(ctr + (depth-rad)*E_3, sqrt(depth*(2*rad - depth)), E_3);

      fill(Neutral());
      S.draw();
    }
}
